(* This program is free software.  See LICENSE for more information *)


(* smt_helper.ml *)
(** contains all the types, exceptions and helper functions used in the parsing process *)


open Smt_sorts;;
open Smt_tables;;
open Smt_helper_logic;;


(** signifies that a boolean connective was attempted to be used with an unaccepted number of arguments.
    i.e (implies x) or (not x y) 
*)
exception Connective_rank_mismatch;;

(** signifies a Validation error
*)
exception Validation_error of string;;


(** type of a variable's id *)
type var_id = string

(** the type of an annotation. *)
type annotation = string * string option

(** contains a list of annotations*)
type annotations = annotation list

(** represents a quantifier *)
type quant =  FORALL 
            | EXISTS

(** represents the name of something *)
type name= NAME of string

(** the type that sorts are stored as *)
type sort_name = string

(** associates a variable id to its corresponding sort*)
type binder = var_id * sort_name

(** the type used to represent a sort declaration along with any annotations that it carries*)
type annotated_sort = ANNOTATED_SORT of sort_name * annotations

(** the type used to represent a predicate declaration, along with any annotations it carries.*)
type pred_sig = PRED_SIG of string * sort_name list * annotations

(** the type used to represent a function declaration, along with any annotations it carries.*)
type func_sig = FUNC_SIG of string * sort_name list * sort_name * annotations

type term =
    T of term2 (** term without annotations*)
  | TA of term2 * annotations (** term with annotations*)
(** the datatype that is used to store a term*)
and term2 =
  | VAR of var_id (** a variable(could be a bound variable, or a variable from an earlier let formula*)
  | NUMERAL of Big_int.big_int (** a numeral *)
  | FUN of string * term list (** A function*)
  | ITE of formula * term * term (** An if then else term.*)
(** the datatype used to store the type of term being used*)
and formula =
    F of formula2 (** a formula without annotations*)
  | FA of formula2 *annotations (** an annotated formula*)
(** the datatype that is used to store a formula *)
and formula2 =
    FLET of string * formula * formula (** a flet formula *)
  | LET of string * term * formula (** a let formula *)
  | FVAR of string (** a variable that corresponds to a previous flet formula*)
  | AND of formula list (** a and b and c and d ... where a b c d ... are formulas*)
  | OR of formula list (** a or b or c or d or ... where a b c d ... are formulas*)
  | NOT of formula (** not a where a is a formula*)
  | IMPLIES of formula*formula (** a-> b where a and b are formulas*)
  | XOR of formula list (** a xor b xor c xor d ... where a b c d ... are formulas*)
  | IFF of formula list (** a <--> b <--> c ... where a b c d ... are formulas*)
  | DISTINCT of term list (** t1 is distinct from t2 is distinct from t3 ... where t1, t2, t3 ,... are terms *)
  | EQUALS of term list (** t1=t2=t3=... where t1 t2 t3 ... are terms*)
  | IF_THEN_ELSE of formula * formula * formula (** Store the formula If Then Else construct *)
  | PRED of string * term list (** Store a predicate with a list of its arguments.  Note that a propositional predicate would simply use the empty list for terms *)
  | TRUE_PRED (** Stores the predicate for true *)
  | FALSE_PRED (** Stores the predicate for false *)
  | QUANT of quant * binder list * formula (** Stores the type of quantifier in a, the list of (variable, sort) pairs in b, and the quantified subformula in c *)
(** the datatype used to store the type of formula being used*)


(** Used to store the status of a benchmark*)
type status = SAT 
            | UNSAT
            | UNKNOWN




(** Used to store a single attribute for a benchmark *)
type benchmark_attribute =
                            ASSUMPTION of formula (** Stores an assumption for a benchmark *)
                          | FORMULA of formula (** Stores a formula to be checked for a benchmark *)
                          | STATUS of status (** stores the status of a benchmark *)
                          | ANNOTATION of annotation (** Stores an annotation for a benchmark*)
                          | LOGIC of name
                          | EXTRASORTS of sort_name list (** stores a list extra sorts that the benchmark uses*)
                          | EXTRAFUNS of func_sig list (** stores a list of extra functions with their signatures for the benchmark *)
                          | EXTRAPREDS of pred_sig list (** stores a list of extra predicates with their signatures for the benchmark *)
                          | NOTES of string (** stores any notes that are relevant to the benchmark *)

(** Stores the name of a benchmark and a list of all the benchmark attributes that it has *)
type benchmark = BENCH of name * benchmark_attribute list



(*************** theory helper functions *****************)

let is_formula a = 
	match a with
	FORMULA b -> true
	| _ -> false;;
	
let is_assumption a = 
	match a with
	ASSUMPTION b -> true
	| _ -> false;;
	
	
let get_benchmark_attributes benchmark = 
	match benchmark with 
	BENCH(a,b) -> b;;

let get_formula attribute = 
	match attribute with
	  FORMULA(formula) -> formula;;

let get_assumption attribute = 
	match attribute with
	ASSUMPTION(formula) -> formula;;

let get_formula_from_benchmark benchmark = 
      List.map get_formula (List.filter is_formula (get_benchmark_attributes benchmark));;  

let get_assumption_from_benchmark benchmark = 
      List.map get_assumption (List.filter is_assumption (get_benchmark_attributes benchmark));;  


	(***************** Well-sortedness functions ****************************)  



let is_sort_of id env_item = 
	match env_item with
		PREDICATE_SORT (name, arg_sorts) -> if (id = name) then true else false
		| FUNCTION_SORT (name, arg_sorts, ret_sort) -> if (id = name) then true else false
		| FORMULA_SORT (name) -> if (id = name) then true else false
		| TERM_SORT (name, sort) -> if (id = name) then true else false
		| VAR_SORT (name, sort) -> if (id = name) then true else false;;

let has_sort id env = 
	if (List.length(List.filter (is_sort_of id) env) = 0)
	then false
	else true;;
	
let get_sort_of id env_item = 
	match env_item with
		PREDICATE_SORT (pname, arg_sorts) -> if (id = pname) then "_boolean_" else raise (Validation_error ("id and env_item don't match in call to get_sort_of "^id^" for predicate named "^pname^"."))
		| FUNCTION_SORT (fname, arg_sorts, ret_sort) -> if (id = fname) then ret_sort else raise (Validation_error ("id and env_item don't match in call to get_sort_of "^id^" for function named "^fname^"."))
		| FORMULA_SORT (fname) -> if (id = fname) then "_boolean_" else raise (Validation_error ("id and env_item don't match in call to get_sort_of "^id^" for formula named "^fname^"."))
		| TERM_SORT (tname, sort) -> if (id = tname) then sort else raise (Validation_error ("id and env_item don't match in call to get_sort_of "^id^" for term named "^tname^"."))
		| VAR_SORT (vname, sort) -> if (id = vname) then sort else raise (Validation_error ("id and env_item don't match in call to get_sort_of "^id^" for var named "^vname^"."));;	


let get_arg_sorts_of id env_item = 
	match env_item with
		PREDICATE_SORT (pname, arg_sorts) -> if (id = pname) then arg_sorts else raise (Validation_error ("id and env_item don't match in call to get_arg_sorts_of "^id^" for predicate named "^pname^"."))
		| FUNCTION_SORT (fname, arg_sorts, ret_sort) -> if (id = fname) then arg_sorts else raise (Validation_error ("id and env_item don't match in call to get_arg_sorts_of "^id^" for function named "^fname^"."))
		| _ -> raise (Validation_error ("get_arg_sorts_of can only be called with a PREDICATE or FUNCTION identifier."));;	

let get_item_from_env id env = 
	List.hd (List.filter (is_sort_of id) env);;


let get_sort id env = 
	if (has_sort id env) then
		List.hd(List.map (get_sort_of id) (List.filter (is_sort_of id) env))
	else
		raise (Validation_error ("sort "^id^" not defined."));;	
	

let add_sort sort env = 
	sort::env;;

let rec get_term_sort env term =
	match term with
		T(term2) -> get_term2_sort env term2
	|	TA(term2, annotations) -> get_term2_sort env term2 
	
and 
	get_term2_sort env term2 =
	match term2 with
		VAR(id) -> 
			let s = get_sort id env in
(*			Printf.printf "\nid: %s , sort: %s" id s;  *)
			s
	|	NUMERAL(num) ->
			let num_string = Big_int.string_of_big_int num in 
			let s = 
			if (has_sort num_string env) then
				get_sort num_string env
			else
				Hashtbl.find Smt_tables.sorts "_numeralsort_"
			in
(*			Printf.printf "\nnum: %s, sort %s" num_string s;  *)
			s
	|	FUN(name, terms) -> 

(* PR: disabled because +, * with arbitrary arity are not correctly handled

		if (List.for_all2 (=) (get_arg_sorts_of name (get_item_from_env name env)) (List.map (get_term_sort env) terms)) then *)
			let s = get_sort name env in
(*			Printf.printf "\nfun: %s , sort: %s" name s;   *)
			s
(*		else
			raise (Validation_error ("Terms of the function named "^name^" do not match the function signature.  Terms:"^(List.fold_left (^) "" (List.map (get_term_sort env) terms))))  *)
	|	ITE(formula, term1, term2) ->
		let x = check_formula env formula in
		let sort1 = get_term_sort env term1 in
		let sort2 = get_term_sort env term2 in
		if (not (sort1 = sort2)) then
		  		raise (Validation_error ("Terms of ITE are of different types:"^ sort1 ^":"^sort2^"."))
		else
			sort1
and
	check_formula env formula =
	  match formula with
	  	F(formula2) -> check_formula2 env formula2
	  |	FA(formula2, annotations) -> check_formula2 env formula2
and
	check_formula2 env formula2 =
	  match formula2 with
	  	FLET(str, f1, f2) -> 
	  	 (* check f1 and add sort of f1 (named by str) to the environment, then check f2 with the new environment *)
	  		check_formula (add_sort (FORMULA_SORT(str)) (check_formula env f1)) f2
	  |	LET(str, term, f1) -> 
	     (* check term and add sort of term (named by str) to the environment, then check f1 with the new environment *)
	  		check_formula (add_sort (TERM_SORT(str,(get_term_sort env term))) env) f1
	  | FVAR(str) ->  
	  	 (* check that the sort of str exists in the current environment *)
	  		if (has_sort str env) then
	  			env
	  		else
	  			raise (Validation_error ("sort "^str^" not defined."))
	  | AND(formulas) -> 
	  	(* check all formulas and return the current environment *)
	  		let x = List.map (check_formula env) formulas in
	  		env
	  | OR(formulas) ->
	  	(* check all formulas and return the current environment *)
	  		let x = List.map (check_formula env) formulas in
	  		env
	  | NOT(formula) -> 
	  	(* check the formula, then return the current environment *)
	  		let x = check_formula env formula in
	  		env
	  | IMPLIES(formula1, formula2) -> 
	  	(* check formula1 and formula2, then return the current environment *)
	  		let x = check_formula env formula1 in
	  		let y = check_formula env formula2 in
	  		env
	  | XOR(formulas) -> 
	  	(* check all formulas and return the current environment *)
	  		let x = List.map (check_formula env) formulas in
	  		env
	  | IFF(formulas) -> 
	  	(* check all formulas and return the current environment *)
	  		let x = List.map (check_formula env) formulas in
	  		env
	  | DISTINCT(terms) -> 
	  	(* check that all the terms are of the same sort, then return the current environment *)	  		
	  		let term_sorts = List.sort String.compare (List.map (get_term_sort env) terms) in
	  		if ((List.hd term_sorts) = (List.hd (List.rev term_sorts))) then
		  		env
	  		else
		  		raise (Validation_error ("DISTINCT:terms are of different types:"^ (List.fold_left (^) "" term_sorts)))
	  | EQUALS(terms) ->
	   (* check that the terms are of the same sort, then return the current environment *)
	  		let term_sorts = List.sort String.compare (List.map (get_term_sort env) terms) in
	  		if ((List.hd term_sorts) = (List.hd (List.rev term_sorts))) then
		  		env
	  		else
		  		raise (Validation_error ("EQUALS:terms are of different types:"^ (List.fold_left (^) "" term_sorts)))
	  | IF_THEN_ELSE(formula1, formula2, formula3) -> 
	  (* check the three formulas, then return the current environment *)
	  		let a = check_formula env formula1 in
	  		let b = check_formula env formula2 in
	  		let c = check_formula env formula3 in
	  		env
	  | PRED(name, terms) -> 
	   (* check the sorts of the terms to make sure they match the predicate signature *)
		if 
		(List.for_all2 (=) (get_arg_sorts_of name (get_item_from_env name env)) (List.map (get_term_sort env) terms)) then
			let s = get_sort name env in
(*			Printf.printf "\npred: %s , sort: %s" name s; *)
			env
		else
			raise (Validation_error ("Terms of the predicate named "^name ^":"^ (List.fold_left (^) "" (List.map (get_term_sort env) terms)) ^ " do not match the predicate signature:"^name ^":"^ (List.fold_left (^) "" (get_arg_sorts_of name (get_item_from_env name env)))))
	  | TRUE_PRED -> env
	  | FALSE_PRED -> env
	  | QUANT(quant, binders, formula) -> 
			let binder_items = 
				(List.map (fun binder -> match binder with (vname, vsort) -> VAR_SORT (vname, vsort)) binders) in
				check_formula (binder_items@env) formula
	  ;; 
	  






